/*
 * labjack - PD External for LabJack Cards
 * 
 * Ecole Nationale Superieure d'Ingenieurs de Bretagne Sud
 * Copyright (C) 2012 Florent de Lamotte
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version. (see LICENCE.txt)
 *
 */

#include "m_pd.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include <pthread.h>
#include <semaphore.h>

#include "labjackusb.h"
#include <libusb-1.0/libusb.h>

#include "labjack.h"
#include "LJ_AISample.h"


// Demonstrates how to build the AIStream packet.
extern void buildAISampleBytes(BYTE * sendBuffer, t_int * ch_config);

// Demonstrates how to parse the response of AIStream.
extern void parseAISampleBytes(BYTE * recBuffer, t_float * ai);


static void *acquire_thread(void *w)
{
    	t_LJ_AISample *x = (t_LJ_AISample*) w;

    	// Setup the variables we will need.
    	int r = 0; // For checking return values
    	BYTE sendBuffer[U12_COMMAND_LENGTH];
    	BYTE recBuffer[U12_COMMAND_LENGTH];

	int i;

    	while(1) {
    		sem_wait(&x->go);

    		// Builds the AISample command
    		buildAISampleBytes(sendBuffer, x->ch_config);
    
    		// Write the command, and read the response.
    		r = writeRead(devHandle, sendBuffer, recBuffer );
    
    		if( r == 0){
    			parseAISampleBytes(recBuffer, x->ai_v);
		}

		// if we have more than 4 channels, read onece more
		if (x->channels > 4) {
    			// Builds the AISample command
    			buildAISampleBytes(sendBuffer, x->ch_config + 4);
   	 
    			// Write the command, and read the response.
    			r = writeRead(devHandle, sendBuffer, recBuffer);
        		if(r == 0){
    				parseAISampleBytes(recBuffer, x->ai_v + 4);
			}
		}

		sem_post(&x->acquired); // acquisition terminee
    		sem_post(&x->busy);     // on libere le thread
    	}

}

static void thread_start(t_LJ_AISample *x)
{
	// create the worker thread
    	if(pthread_attr_init(&x->LJ_AISample_thread_attr) < 0){
		sys_lock();
       		error("labjack: could not launch acquire thread");
	   	sys_unlock();
       		return;
    	} else if(pthread_attr_setdetachstate(&x->LJ_AISample_thread_attr, PTHREAD_CREATE_DETACHED) < 0) {
       		sys_lock();
	   	error("labjack: could not launch acquire thread");
       		sys_unlock();
	   	return;
    	}

    	if(pthread_create(&x->x_threadid, &x->LJ_AISample_thread_attr, acquire_thread, x) < 0) {
       		sys_lock();
	   	error("labjack: could not launch aquire thread");
	   	sys_unlock();
       		return;
    	} 
}


/*
 * Executed when banged ... take a shot ;)
 */
void LJ_AISample_sample(t_LJ_AISample *x)
{
	if (sem_trywait(&(x->busy)) != -1) {	
		sem_post(&x->go);
  		clock_delay (x->x_clock, 10);
	} else {
		outlet_bang(x->err);
	}
}

/*
 * After thread is released, periodically check if outputs are updated
 */
static void LJ_AISample_update(t_LJ_AISample *x)
{
	int i;
	if (sem_trywait(&x->acquired) != -1) {
		for (i = 0; i < x->channels; i++) {
			outlet_float(x->ai[i], x->ai_v[i]);
		}
/*
		for (i = 0; i < 4; i++) {
			outlet_float(x->io[i], x->io_v[i]);
		}
*/
	} else { // if data not ready, wait another 2ms
		clock_delay (x->x_clock, 2);
	}
}


void *LJ_AISample_new(t_symbol *s, int argc, t_atom *argv)
{
	t_LJ_AISample *x;	
	x = (t_LJ_AISample *)pd_new(LJ_AISample_class);
	int i;
	
	// one outlet for each analog input
	x->channels = argc;
	for (i = 0; i < argc; i++) {
		x->ai[i] = outlet_new(&x->p_ob, &s_float);
		x->ch_config[i] = atom_getfloat(argv+i);		
	}
	for (i = argc; i<8; i++) x->ch_config[i] = 0;

/*
	for (i = 0; i < 4; i++) {
		x->io[i] = outlet_new(&x->p_ob, &s_float);
	}
*/

	x->err = outlet_new(&x->p_ob, &s_bang);
	x->errors = 0;

	// a clock to update outputs 
  	x->x_clock = clock_new(x, (t_method) LJ_AISample_update);

	// setting up thread and semaphores
	sem_init (&x->busy, 0, 1);
	sem_init (&x->go, 0, 0);
	sem_init (&x->acquired, 0, 0);

	thread_start(x);

	return x; 
}

void LJ_AISample_free(t_LJ_AISample *x)
{
	while(pthread_cancel(x->x_threadid) < 0);	
}

// Uses information from section 5.1 of the U12 User's Guide to make a AISample
// packet.
// http://labjack.com/support/u12/users-guide/5.1
void buildAISampleBytes(BYTE * sendBuffer, t_int * ch_config) {
    // Build up the bytes
    sendBuffer[0] = 8 + ch_config[0];   // Set PGAMUX 
    sendBuffer[1] = 8 + ch_config[1];   // Set PGAMUX
    sendBuffer[2] = 8 + ch_config[2];   // Set PGAMUX
    sendBuffer[3] = 8 + ch_config[3];   // Set PGAMUX
    sendBuffer[4] = 1;   // UpdateIO = 0, LEDState = 1
    sendBuffer[5] = 192; // 0b11000000 = (AISample)
    sendBuffer[6] = 0;   // XXXXXXXX
    sendBuffer[7] = 0;   // Echo Value
    
    // The bytes have been set. We are ready to write to the U12.
}

// Parses the AISample packet into something useful.
void parseAISampleBytes(BYTE * recBuffer, t_float * ai){
    int temp;
    
    // Apply the single-ended conversion to results
    temp = (recBuffer[2] >> 4) & 0xf;
    temp = (temp << 8) + recBuffer[3];
    ai[0] = ((double)temp * 20.0 / 4096.0) - 10;
    
    temp = recBuffer[2] & 0xf;
    temp = (temp << 8) + recBuffer[4];
    ai[1] = ((double)temp * 20.0 / 4096.0) - 10;
    
    temp = (recBuffer[5] >> 4) & 0xf;
    temp = (temp << 8) + recBuffer[6];
    ai[2] = ((double)temp * 20.0 / 4096.0) - 10;
    
    temp = recBuffer[5] & 0xf;
    temp = (temp << 8) + recBuffer[7];
    ai[3] = ((double)temp * 20.0 / 4096.0) - 10;



/*
    printf("Results of AISample:\n");
    printf("  AI0 = %f\n", ai0);
    printf("  AI1 = %f\n", ai1);
    printf("  AI2 = %f\n", ai2);
    printf("  AI3 = %f\n", ai3);
    printf("  PGA Overvoltage = %d\n", (recBuffer[0] >> 4) & 1);
    printf("  IO3 to IO0 States = %d\n", recBuffer[0] & 15);
    printf("  TimerCounterMask = %d\n", recBuffer[22]);
*/  
}
